from lhcsmapi.dbsignal.Signal import Signal
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities

import pandas as pd
import numpy as np
import datetime
from lhcsmapi.Time import Time


def validate_key_in_kwargs(key, *key_types, **kwargs):
    result = None
    if key in kwargs:
        result = kwargs[key]
        match_found = False
        for key_type in key_types:
            match_found = match_found or isinstance(result, key_type)
        if not match_found:
            raise TypeError("You entered a {} keyword, but the keyword needs to be a {}.".format(key, key_types))

    return result


class SignalQuery(object):
    # use property
    def set_db(self, db):
        self.db = db

    def set_circuit_type(self, circuit_type):
        self.circuit_type = circuit_type

    def set_t_start(self, t_start):
        self.t_start = t_start

    def set_t_end(self, t_end):
        self.t_end = t_end

    def set_signal(self, signal):
        self.signal = signal

    def set_metadata(self, metadata):
        self.metadata = metadata

    def __str__(self):
        return f'Querying {self.db}, for circuit  {self.circuit_type}, from {self.t_start} to {self.t_end} for {self.signal}'

    def __repr__(self):
        return f'Querying {self.db}, for circuit  {self.circuit_type}, from {self.t_start} to {self.t_end} for {self.signal}'


class SignalQueryBuilder(object):

    def __init__(self):
        self._query = SignalQuery()

    def __getattr__(self, name):
        if name not in ['_query', 'with_db']:
            raise AttributeError(
                '{}() is not supported, you can only call with_db(), e.g., with_db(\'NXCALS\')'.format(name))

    def __repr__(self):
        return 'Set database name using with_db() method, e.g., with_db(\'NXCALS\')'

    def with_db(self, db):
        # Check if db is supported
        self._query.set_db(db)
        return SignalQueryBuilderWithDb(self._query)


class SignalQueryBuilderWithDb(object):

    def __init__(self, query):
        self._query = query

    # Can be generalized
    def __getattr__(self, name):
        if name not in ['_query', 'with_circuit_type']:
            raise AttributeError(
                name + '() is not supported, you can only call with_circuit_type(), e.g., with_circuit_type(\'RB\')')

    def __repr__(self):
        return 'Database name properly set to {}. Set circuit type using with_circuit_type() method, e.g., with_circuit_type(\'RB\').'.format(
            self._query.db)

    def with_circuit_type(self, circuit_type):
        # Check if circuit_type is supported
        self._query.set_circuit_type(circuit_type)
        return SignalQueryBuilderWithDbCircuitType(self._query)


class SignalQueryBuilderWithDbCircuitType(object):

    def __init__(self, query):
        self._query = query

    # Can be generalized
    def __getattr__(self, name):
        if name not in ['_query', 'with_duration']:
            raise AttributeError(
                name + '() is not supported, you can only call with_duration(), e.g., with_duration(t_start=\'2015-01-13 16:59:11+01:00\', t_end=\'2015-01-13 17:15:46+01:00\')')

    def __repr__(self):
        return 'Circuit type properly set to {}. Set duration using with_duration() method, e.g., with_duration(t_start=\'2015-01-13 16:59:11+01:00\', t_end=\'2015-01-13 17:15:46+01:00\').'.format(
            self._query.circuit_type)

    def with_duration(self, **duration):
        if 'eventTime' in duration and self._query.db == 'PM':
            t_start = duration['eventTime']
            t_end = -1
        elif self._query.db in ['NXCALS', 'CALS']:
            t_start = validate_key_in_kwargs('t_start', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64,
                                             **duration) or 0
            t_end = validate_key_in_kwargs('t_end', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64,
                                           **duration)
            duration = validate_key_in_kwargs('duration', tuple, list, **duration)

            t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, '')
        else:
            raise ValueError('Wrong input parameters!')

        self._query.set_t_start(t_start)
        self._query.set_t_end(t_end)

        return SignalQueryBuilderWithDbCircuitTypeDuration(self._query)


class SignalQueryBuilderWithDbCircuitTypeDuration(object):

    def __init__(self, query):
        self._query = query

    # Can be generalized
    def __getattr__(self, name):
        if name not in ['_query', 'with_metadata']:
            raise AttributeError(
                name + '() is not supported, you can only call with_metadata(), e.g., with_metadata(circuit_name=\'RB.A12\', system=\'PC\', signal=\'I_MEAS\')')

    def __repr__(self):
        return 'Circuit duration set to t_start={}, t_end={}. Set metadata using with_metadata() method, e.g., with_metadata(circuit_name=\'RB.A12\', system=\'PC\', signal=\'I_MEAS\').'.format(
            self.query.t_start, self.query.t_end, self.query.circuit_type)

    def with_metadata(self, **md):
        # Take the metadata according to date

        # Check if metadata definition is complete
        circuit_type = self._query.circuit_type
        db = self._query.db

        metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, md['circuit_name'], md['system'],
                                                                       db)
        signal = SignalMetadata.get_signal_name(circuit_type, md['circuit_name'], md['system'], db, md['signal'])

        self._query.set_metadata(metadata)
        self._query.set_signal(signal)
        return SignalQueryBuilderWithDbCircuitTypeDurationMetadata(self._query)


class SignalQueryBuilderWithDbCircuitTypeDurationMetadata(object):

    def __init__(self, query):
        self._query = query

    # Can be generalized
    def __getattr__(self, name):
        if name not in ['_query', 'query']:
            raise AttributeError('{}() is not supported, you can only call query(), e.g., query(spark)'.format(name))

    def __repr__(self):
        return 'Start time set to {} and end time to {}. Query signal using query() method, e.g., query().'.format(
            self._query.t_start, self._query.t_end, self._query.circuit_type)

    def query(self, dbconnector=None):

        # For NXCALS and PM
        metadata = self._query.metadata
        metadata_db = {k: v for k, v in metadata.items() if k[0].islower()}

        additional_params = {'signal': self._query.signal}
        # Set time - this should be in get_time method
        if self._query.db == 'PM':
            additional_params['eventTime'] = self._query.t_start
        elif self._query.db in ['NXCALS', 'CALS']:
            additional_params['t_start'] = self._query.t_end
            additional_params['t_end'] = self._query.t_end

        # Set additional metadata - this should be in get metadata method
        if self._query.db == 'PM':
            metadata_db = {k: v for k, v in metadata_db.items() if k[0].islower()}
        if self._query.db == 'NXCALS':
            metadata_db = {'nxcals_{}'.format(k): v for k, v in metadata_db.items() if k[0].islower()}
            additional_params['spark'] = dbconnector
        elif self._query.db == 'CALS':
            metadata_bb = {}
            additional_params['ldb'] = dbconnector

        params = {**metadata_db, **additional_params}
        df = Signal().read(self._query.db, **params)

        return SignalProcessing(df, self._query.db)


class SignalProcessing(object):

    def __init__(self, df, db):
        self.df = df
        self.db = db

    def __getattr__(self, name):
        method_list = [func for func in dir(SignalProcessing) if
                       callable(getattr(SignalProcessing, func)) and not func.startswith("__")]
        if name not in method_list:
            raise AttributeError(name + '{}() is not supported, you can call {}'.format(name, method_list))

    def __repr__(self):
        # Check available method names
        method_list = [func for func in dir(SignalProcessing) if
                       callable(getattr(SignalProcessing, func)) and not func.startswith("__")]
        return 'Available methods are {}'.format(method_list)

    def synchronize_time(self, t0=None):
        self.df = SignalUtilities.synchronize_df(self.df, t0)
        return self

    def convert_index_to_sec(self):
        self.df = SignalUtilities.convert_indices_to_sec(self.df)
        return self

    def create_col_from_index(self, col):
        self.df[col] = self.index
        return self

    def filter_median(self, window):
        self.df = self.df.rolling(window=3, min_periods=1).median()
        return self

spark=1
i_meas_nxcals_df = SignalQueryBuilder().with_db('NXCALS').with_circuit_type('RB')\
    .with_duration(t_start='2015-01-13 16:59:11+01:00', t_end='2015-01-13 17:15:46+01:00')\
    .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').query(spark)